/* PACE Project, Rice University
 *
 * This code measures the parameters of the cache memory system.
 * 
 * First, it determines L1 size, associativity, and line size.
 * It uses size to determine associativity.
 *
 *
 */

#include "MemTest.h"
#include "BlackJackTimer.h"

void exit( int x );
double sqrt(double x);
int errno;



void shuffle( int array[], int size )
{
  int i, j, t;

  for (i=0; i<size; i++)
    array[i] = i;

  for (i=0;i<size;i++)
    {
      j        = rand() % size;
      t        = array[i];
      array[i] = array[j];
      array[j] = t;
    }
}

void PrintArray( char *s, int array[], int size )
{
  int i;

  fprintf(stderr,"\n\nContents of %s: (%d)\n",s,size);
  for (i=0;i<size;i++)
    {
      if ((i % 4) == 0)
	fprintf(stderr,"\n%10d:  ",i);
      fprintf(stderr,"%10d, ",array[i]);
    }
  fprintf(stderr,"\n");
}

static void DieWithMessage(char *currArg, int argNumber)
{
  fprintf(stderr,"Did not understand argument '%s' (%d).\n",
	  currArg,argNumber);
  fprintf(stderr,"Valid arguments are:\n");
  fprintf(stderr,"\t-g \tActivate internal debugging checks.\n");
  fprintf(stderr,"\t-l <name> \tUse <name> for log file.\n");
  fprintf(stderr,"\t-v \tPrint results (& debug info) incrementally.\n");
  exit(-1);  
}

int main(int argc, char *argv[])
{
  int i, j, arg, argNumber;
  char *currArg, *LogFileName;

  int    Sizes[MAX_TESTS_PER_RUN];
  double Times[MAX_TESTS_PER_RUN];
  int    Cycles[MAX_TESTS_PER_RUN];
  int    Count;

  Verbose   = 0;
  Debug     = 0;
  HeartBeat = 1;

  LogFileName = "./Log";

  /* Parse the command line parameters */
  for (argNumber = 1; argNumber < argc; argNumber++)
  {
    currArg = argv[argNumber];

    if (*currArg == '-')
    {
      currArg++; 
      while (*currArg != '\0')
      {
	switch(*currArg)
        {
	  case 'd':
	    Debug++;
	    break;
	  case 'g':
	    Debug++;
	    break;
  	  case 'h':
	    HeartBeat++;
	    break;
	  case 'l':
	    argNumber++;
	    LogFileName = argv[argNumber];
	    break;
	  case 'v':
            Verbose++;
	    break;
	  default:
	    DieWithMessage(currArg,argNumber);
	}
	currArg++;
      }
    }
  }

  /* Set up the Log File */
  fprintf(stderr,"Log File Name is '%s'.\n",LogFileName);

  LogFile = fopen(LogFileName,"w");
  if (LogFile == NULL) /* cannot use abort() since it writes 'LogFile' */
  {
    fprintf(stderr,"fopen() of 'CacheTestLog' failed\n");
    exit(-2);
  }

  SetupAndMark( LogFile, "LogFile" );
  fprintf(LogFile,"\"L1 Microbenchmark\"\n");

  if (HeartBeat)
    fprintf(stderr,"Providing HeartBeat @ Level %d.\n",HeartBeat);

  /* test to see if the environment variable for the temporary directory */
  /* has been set.  Aborts if it fails */
  (void) TestEnvVar();  

  /* Some initializations */
  UnitSize = sizeof(void **);          /* tests use a chain of 'void *' s */

  /* Find an appropriate timer interval to allow 0.5% accuracy */
  TimerTick = FindTimerTick();
  MinTime = MIN_TICKS * TimerTick;

  fprintf(LogFile,"\"Timer tick is %s, Minimum time is %s\"\n",
	  PrintDNum(TimerTick), PrintDNum(MinTime));

  /* Initializations */
  BigInt = FindBigInt();
  i = (int) fmod(getticks(),1000000.0);
  fprintf(stderr,"Random seed is %s.\n",PrintNum(i));
  srand( i ); /* no real reason to believe in a better seed */

  AddCostInNSec = DReadResult("AddCost");
  fprintf(LogFile,"Add cost is %f.\n",AddCostInNSec);

  int FindL1 = 1;
  if (FindL1)  /* assumes that L1 < 1 MB */
  {
    arg = GenerateLogSet(Sizes, MAX_TESTS_PER_RUN, 
			 LOWERBOUNDINBYTES/UnitSize, 
			 UPPERBOUNDINBYTES/UnitSize);
    i = Min(512,LOWERBOUNDINBYTES/UnitSize);
    BCTrial(Sizes, Times, arg, i, 64);
    AnalyzeL1(Sizes, Times, arg, 10*TimerTick);
  }

  i = PACE_AllocationCheck();
  if (i != 0)
    fprintf(stderr,"Memory leak claimed %d aligned items.\n",i);
}
  
